import java.io.*;
import java.util.ArrayList;
import javax.swing.JOptionPane;
public class GillespieSolver
{
	private String project;
	private int nSpecies;
	private int nReactions;
	private int[][] changes;
	private int[][] lefthand;
	private int[][] righthand;
	private int maxReactions = 100;
	private ArrayList<String> speciesNames;
	private int[] reactionType;
	private int[][] reactionComp;
	private int[] population;
	private double[] rates;
	private double t;
	public GillespieSolver(String pD)
	{
		project = pD;
		initialise();
	}
	
	
	private void initialise()
	{
		t = 0.0;
		
		String speciesFile = project + "/species.txt";
		System.out.println(String.format("Loading %s",speciesFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(speciesFile)); 
			speciesNames = new ArrayList<String>();
			String line = "";
			while((line = in.readLine())!=null)
			{
				speciesNames.add(line);	
			}
			nSpecies = speciesNames.size();
			System.out.println(String.format("Read %d species",nSpecies));
			in.close();
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",speciesFile));	
			System.exit(0);
		}
		
		
		/*String changeFile = project + "/changes.txt";
		System.out.println(String.format("Loading %s",changeFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(changeFile)); 
			String line = "";
			changes = new int[maxReactions][nSpecies];
			int pos = 0;
			while((line = in.readLine())!=null)
			{
				String[] splitLine = line.split("\t");
				for(int i=0;i<nSpecies;i++)
				{
					changes[pos][i] = Integer.parseInt(splitLine[i]);	
				}	
				pos++;
			}
			nReactions = pos;
			
			in.close();
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",changeFile));	
			System.exit(0);
		}
		*/
		String reactionFile = project + "/reactions.txt";
		System.out.println(String.format("Loading %s",reactionFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(reactionFile)); 
			String line = "";
			reactionType = new int[nReactions];
			lefthand = new int[maxReactions][nSpecies];
			righthand = new int[maxReactions][nSpecies];
			rates = new double[maxReactions];
			reactionType = new int[maxReactions];
			int pos = 0;
			while((line = in.readLine())!=null)
			{
				System.out.println(line);
				String[] components = line.split(":");
				rates[pos] = Double.parseDouble(components[3]);
				reactionType[pos] = Integer.parseInt(components[0]);
				if(components[1].length()>0)
				{
					String[] left = components[1].split(",");
					for(int i=0;i<left.length;i++)
					{
						int ind = speciesNames.indexOf(left[i]);
						lefthand[pos][ind] += 1;	
					}	
				}
				if(components[2].length()>0)
				{
					String[] right = components[2].split(",");
					for(int i=0;i<right.length;i++)
					{
						int ind = speciesNames.indexOf(right[i]);
						righthand[pos][ind] += 1;	
					}	
				}
				pos ++;	
			}				
			nReactions = pos;
			System.out.println(String.format("Read %d reactions",nReactions));
			for(int i=0;i<nReactions;i++)
			{
				String temp = "";
				for(int j=0;j<nSpecies;j++)
				{
					temp += " " + lefthand[i][j];	
				}
				temp += " : ";
				for(int j=0;j<nSpecies;j++)
				{
					temp += " " +righthand[i][j];	
				}
				System.out.println(temp);				
			}
			for(int i=0;i<nReactions;i++)
			{
				System.out.println(rates[i]);
			}		

			in.close();
			changes = new int[maxReactions][nSpecies];
			for(int i=0;i<nReactions;i++)
			{
				for(int j=0;j<nSpecies;j++)
				{
					changes[i][j] = righthand[i][j] - lefthand[i][j];	
				}	
			}
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",reactionFile));	
			System.exit(0);
		}
		
		
		String initFile = project + "/init.txt";
		System.out.println(String.format("Loading %s",initFile));
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(initFile)); 
			String line = "";
			population = new int[nSpecies];
			for(int i=0;i<nSpecies;i++)
			{
				line = in.readLine();
				population[i] = Integer.parseInt(line);
			}
			/*for(int i=0;i<nReactions;i++)
			{
				line = in.readLine();
				rates[i] = Double.parseDouble(line);	
			}*/
			System.out.println(String.format("Loaded %s:",initFile));
			System.out.println("Species:");
			for(int i=0;i<nSpecies;i++)
			{
				System.out.println(String.format("%s:%d",speciesNames.get(i),population[i]));	
			}
			/*System.out.println("Rates:");
			for(int i=0;i<nSpecies;i++)
			{
				System.out.println(String.format("%d:%f",i,rates[i]));	
			}*/
			in.close();
		}
		catch(IOException e)
		{
			System.out.println(String.format("Failed to load %s",initFile));	
			System.exit(0);	
		}
		
		
			
	}
	
	private double compA(int reactionNo)
	{
		double tempa = rates[reactionNo];
		for(int i=0;i<nSpecies;i++)
		{
			if(lefthand[reactionNo][i]==1)
			{
				tempa *= population[i];	
			}	
			else if(lefthand[reactionNo][i]==2)
			{
				tempa *= 0.5*population[i]*(population[i]-1.0);
			}
		}
/*		if(reactionType[reactionNo]==0)
		{
			for(int i=0;i<nSpecies;i++)
			{
				if(reactionComp[reactionNo][i]==1)
				{
					tempa *= population[i];
				}
			}	
			
		}else if(reactionType[reactionNo]==1)
		{
			for(int i=0;i<nSpecies;i++)
			{
				if(reactionComp[reactionNo][i]==1)
				{
					tempa *= 0.5*(population[i]*(population[i]-1));	
				}	
			}
		}*/
		return tempa;	
	}
	public void setPopulation(int[] popin)
	{
		for(int i=0;i<nSpecies;i++)
		{
			population[i] = popin[i];	
		}	
	}
	public int[] getPopulation()
	{
		return population;	
	}
	public int getnSpecies()
	{
		return nSpecies;	
	}
	public int runMulti(double maxT,int nRep,String outName)
	{
		PrintWriter out;
		System.out.println("Running");
		int totReactions = 0;
		try
		{
			int[] temp = new int[nSpecies];
			for(int i=0;i<nSpecies;i++)
			{
				temp[i] = population[i];	
			}
			
			out = new PrintWriter(new FileWriter(outName)); 
			String head = "It";
			for(int i=0;i<nSpecies;i++)
			{
				head += String.format("\t%s",speciesNames.get(i));	
			}
			out.println(head);
			for(int rep = 0;rep<nRep;rep++)
			{
				// Set the population
				for(int i=0;i<nSpecies;i++)
				{
					population[i] = temp[i];
				}
				t = 0.0;
				totReactions += run(maxT,null);
				head = String.format("%d",rep);
				for(int i=0;i<nSpecies;i++)
				{
					head += String.format("\t%d",population[i]);	
				}
				out.println(head);
			}
			out.close();
		}
		catch(IOException e)
		{
			System.out.println("Failed to open output file");
			System.exit(0);	
		}
		return totReactions;

	} 
	public int run(double maxT,String outName)
	{
		int totReactions = 0;
		PrintWriter out;
		System.out.println("Running");
		try
		{
			if(outName!=null)
			{
				out = new PrintWriter(new FileWriter(outName)); 
			}
			else
			{
				out = null;
			}
			String head = "T";
			for(int i=0;i<nSpecies;i++)
			{
				head += String.format("\t%s",speciesNames.get(i));	
			}
			if(out!=null)
			{
				out.println(head);
			}
			String outline = String.format("%f",t);
			for(int i=0;i<nSpecies;i++)
			{
				outline += String.format("\t%d",population[i]);	
			}
			if(out!=null)
			{
				out.println(outline);
			}
			double[] a = new double[nReactions];
			while(t<maxT)
			{
				double a0 = 0.0;
				for(int i=0;i<nReactions;i++)
				{
					a[i] = compA(i);
					a0+=a[i];	
				}
				if(a0==0)
				{
					break;
				}
				// Sample time to next reaction
				double tau = (1/a0)*Math.log(1/Math.random());
				t += tau;
				if(t>maxT)
				{
					break;
				}
				outline = String.format("%f",t);
				
				int pos = 0;
					
				double r = Math.random();
				double c = a[0]/a0;
				while(r>c)
				{
					pos++;
					c = c + a[pos]/a0;
				}
				
				// Update population
				for(int i=0;i<nSpecies;i++)
				{
					population[i] += changes[pos][i];	
					outline += String.format("\t%d",population[i]);
				}
				if(out!=null)
				{
					out.println(outline);				
				}
				totReactions ++;
			}
			System.out.println(String.format("Finished after %d reactions",totReactions));
			if(out!=null)
			{
				out.close();
			}
		}
		catch(IOException e)
		{
			System.out.println("Failed to open output file");
			System.exit(0);	
		}
		return totReactions;

	}
	
	public static void main(String[] args)
	{
		String projectName = "";
		double maxT = 0.0;
		String outName = "";
		int nReps = 0;
		if(args.length==0)
		{
			projectName = (String)JOptionPane.showInputDialog(null,"Enter Project Directory");
			if(projectName==null)
			{
				System.exit(0);	
			}

			while(true)
			{
				String temp = (String)JOptionPane.showInputDialog(null,"Enter maximum simulation time");
				if(temp==null)
				{
					System.exit(0);	
				}
				else
				{
					try
					{
						maxT = Double.parseDouble(temp);	
					}
					catch(NumberFormatException e)
					{
						continue;
					}
					break;
				}
			}
			
			outName = (String)JOptionPane.showInputDialog(null,"Enter output file name");
			if(outName == null)
			{
				System.exit(0);	
			}
			
			while(true)
			{
				String temp = (String)JOptionPane.showInputDialog(null,"Enter number of repetitions");
				if(temp==null)
				{
					System.exit(0);	
				}
				else
				{
					try
					{
						nReps = Integer.parseInt(temp);	
					}
					catch(NumberFormatException e)
					{
						continue;
					}
					break;
				}
			}

			
		}
		else if(args.length==4)
		{
			projectName = args[0];	
			try
			{
				maxT = Double.parseDouble(args[1]);
				nReps = Integer.parseInt(args[3]);
			}
			catch(NumberFormatException e)
			{
				System.out.println("Arguments = projectDir 	maxT outName nReps");
				System.exit(0);
			}
			outName = args[2];
		}
		GillespieSolver gs = new GillespieSolver(projectName);
		if(nReps==1)
		{
			//run once and output to file
			int r = gs.run(maxT,outName);
		}
		else
		{
			int r = gs.runMulti(maxT,nReps,outName);			
		}	
	}
		
}